@@ -20,7 +20,7 @@
 
 #include "mozplugger.h"
 
-#define THIS (*(data_t **) &(instance->pdata))
+#define THIS ((data_t *)(instance->pdata))
 
 #ifndef __GNUC__
 #define __inline
@@ -44,8 +44,8 @@ typedef struct data
      int fd;
      int repeats;
      int cmd_flags;
-     char *command;
-     char *winname;
+     const char *command;
+     const char *winname;
      int mode_flags;
      char *mimetype;
      char *href;
@@ -58,14 +58,15 @@ typedef struct data
 
 typedef struct mimetype
 {
-     char type[MAX_TYPE_SIZE];
+     const char * type;
 } mimetype_t;
 
 typedef struct command
 {
      int flags;
-     char cmd[MAX_CMD_SIZE];
-     char winname[MAX_WINNAME_SIZE];
+     const char * cmd;
+     const char * winname;
+     const char * fmatchStr;
 } command_t;
 
 typedef struct handle
@@ -78,7 +79,7 @@ typedef struct handle
 
 typedef struct
 {
-     char *name;
+     const char *name;
      int value;
 } flag_t;
 
@@ -292,7 +293,7 @@ static void run(NPP instance, const char *file, int fd)
 /*****************************************************************************
  * Check if 'file' is somewhere to be found in the user PATH.
  *****************************************************************************/
-static int inpath(char *path, char *file)
+static int inpath(char *path, const char *file)
 {
      static struct stat filestat;
      static char buf[1024];
@@ -324,37 +325,82 @@ static int inpath(char *path, char *file)
      return 0;
 }
 
-/*****************************************************************************
- * Check if 'file' exists.
- *****************************************************************************/
-static int find(char *file)
+typedef struct 
 {
-     static struct stat filestat;
-     static char path[1024];
-     char *env_path;
+    char name[SMALL_BUFFER_SIZE];
+    short exists;
+} cacheEntry_t;
+
+/*****************************************************************************
+ * Check if 'file' exists. Uses a cache of previous finds to avoid performing
+ * a stat call on the same file over and over again.
+ *****************************************************************************/
+static int find(const char *file)
+{
+     static cacheEntry_t cache[FIND_CACHE_SIZE];
+     static int cacheSize = 0;
+     static int cacheHead = 0;
+
+     struct stat filestat;
+     int i;
+     int exists;
 
      D("find(%s)\n", file);
 
-     if (file[0] == '/')
-	  return !stat(file, &filestat);
-     
-     /* Get the environment variable PATH */
-     if (!(env_path = getenv("PATH")))
+     for(i = 0; i < cacheSize; i++)
      {
-	  D("No PATH !\n");
-	  return 0;
+         if( strcmp(cache[i].name, file) == 0)
+         {
+             const int exists = cache[i].exists;
+             D("find cache hit exists = %s\n", exists ? "yes" : "no");
+             return exists;
+         }
      }
-     
-     /* Duplicate to avoid corrupting Mozilla's PATH */
-     strncpy(path, env_path, sizeof(path));
 
-     return inpath(path, file);
+     if (file[0] == '/')
+     {
+	  exists = !stat(file, &filestat);
+     }
+     else
+     {
+          char *path;
+          char *env_path;
+
+          /* Get the environment variable PATH */
+          if (!(env_path = getenv("PATH")))
+          {
+	       D("No PATH !\n");
+	       exists = 0;
+          }
+          else
+          {
+               /* Duplicate to avoid corrupting Mozilla's PATH */
+               path = strdup(env_path);
+               exists = inpath(path, file);
+               free(path);
+          }
+     }
+
+     strncpy(cache[cacheHead].name, file, SMALL_BUFFER_SIZE);
+     cache[cacheHead].name[SMALL_BUFFER_SIZE-1] = '\0';
+     cache[cacheHead].exists = exists;
+
+     cacheHead++;
+     if(cacheHead > cacheSize)
+     {
+          cacheSize = cacheHead;
+     }
+     if(cacheHead >= FIND_CACHE_SIZE)      /* Wrap around the head */
+     {
+          cacheHead = 0;
+     }
+     return exists;
 }
 
 /*****************************************************************************
  * Delete a mime handler if no command has been found for it.
  *****************************************************************************/
-static void filter_previous_handler()
+static void filter_previous_handler(void)
 {
      handler_t *h;
      if (num_handlers >= 1)
@@ -369,19 +415,62 @@ static void filter_previous_handler()
      }
 }
 
+static char staticPool[MAX_STATIC_MEMORY_POOL];
+static int  staticPoolIdx = 0;
+
 /*****************************************************************************
- * Match a word to a flag of the type swallow(name).
+ * Allocate some memory from the static pool. We use a static pool for
+ * the database because Mozilla can unload the plugin after use and this
+ * would lead to memory leaks if we use the heap.
  *****************************************************************************/
-static char *get_winname(char *x, char *flag, command_t *c)
+static void * allocStaticMem(int size)
+{
+     void * retVal; 
+     const int newIdx = staticPoolIdx + size;
+
+     if(newIdx > MAX_STATIC_MEMORY_POOL)
+     {
+	  fprintf(stderr, "MozPlugger: Too many entries in mozpluggerc\n");
+	  exit(1);
+     }
+
+     retVal = &staticPool[staticPoolIdx];
+     staticPoolIdx = newIdx;
+     return retVal;
+}
+
+/*****************************************************************************
+ * Given a pointer to a string in temporary memory, return the same string
+ * but this time stored in permanent (i.e. static) memory. Will only be deleted
+ * when the plugin is unloaded by Mozilla.
+ *****************************************************************************/
+static const char * makeStrStatic(const char * str, int len)
+{
+    /* plus one for string terminator */
+    char * const buf = allocStaticMem(len + 1);
+
+    strncpy(buf, str, len);
+    buf[len] ='\0';
+    return (const char *) buf;
+}
+
+
+/*****************************************************************************
+ * Match a word to a flag that takes a parameter e.g. swallow(name).
+ *****************************************************************************/
+static char *get_parameter(char *x, const char *flag, const char **c)
 {
      char *end;
      int len;
  
-     x += strlen(flag);
-     while (isspace(*x)) x++;
+     /* skip spaces or tabs between flag name and parameter */
+     while ((*x == ' ' || *x == '\t')) 
+         x++;
+
      if (*x != '(')
      {
-	  fprintf(stderr, "MozPlugger: Warning: Expected '(' after '%s'\n", flag);
+	  fprintf(stderr, 
+                  "MozPlugger: Warning: Expected '(' after '%s'\n", flag);
 	  return NULL;
      }
      x++;
@@ -389,13 +478,7 @@ static char *get_winname(char *x, char *flag, command_t *c)
      if (end)
      {
 	  len = end - x;
-	  if (!(len < sizeof(c->winname)))
-	  {
-	       fprintf(stderr, "MozPlugger: Error: Window name too long (%s)\n", x);
-	       exit(1);
-	  }
-	  memcpy(c->winname, x, len);
-	  c->winname[len] = 0;
+	  *c = makeStrStatic(x, len);
 	  x = end+1;
      }
      return x;
@@ -405,7 +488,7 @@ static char *get_winname(char *x, char *flag, command_t *c)
  * Match two words.
  *****************************************************************************/
 __inline
-static int match_word(char *line, char *kw)
+static int match_word(const char *line, const char *kw)
 {
      return !strncasecmp(line, kw, strlen(kw)) && !isalnum(line[strlen(kw)]);
 }
@@ -430,6 +513,8 @@ static int parse_flags(char **x, command_t *commandp)
           { "links",            H_LINKS         },
 	  { "hidden",           H_HIDDEN        },
           { "controls",         H_CONTROLS      },
+          { "swallow",          H_SWALLOW       },
+          { "fmatch",           H_FMATCH        },
 	  { NULL, 		0 		}
      };
      flag_t *f;
@@ -440,13 +525,20 @@ static int parse_flags(char **x, command_t *commandp)
 	  {
 	       *x += strlen(f->name);
 	       commandp->flags |= f->value;
-	       return 1;
-	  }
-	  if (match_word(*x, "swallow"))
-	  {
-	       commandp->flags |= H_SWALLOW;
-	       if ((*x = get_winname(*x, "swallow", commandp)))
-		    return 1;
+
+               /* Get parameter for thoses commands with parameters */
+               if(f->value & H_SWALLOW)
+               {
+	            if ((*x = get_parameter(*x, f->name, &commandp->winname)))
+                        return 1;
+               }
+               else if(f->value & H_FMATCH)
+               {
+	            if ((*x = get_parameter(*x, f->name, &commandp->fmatchStr)))
+                        return 1;
+               }
+               else
+	            return 1;
 	  }
      }
      return 0;
@@ -509,14 +601,7 @@ static void read_config(FILE *f)
 			    handler->num_types, num_handlers, handler->types[0].type);
 		    exit(1);
 	       }
-
-	       if (!(strlen(buffer) < sizeof(type->type)))
-	       {
-		    fprintf(stderr, "MozPlugger: Error: Mime type too long (%s)\n", buffer);
-		    exit(1);
-	       }
-
-	       strncpy(type->type, buffer, sizeof(type->type));
+	       type->type = makeStrStatic(buffer, strlen(buffer));
 	  }
 	  else
 	  {
@@ -564,13 +649,7 @@ static void read_config(FILE *f)
 		    if (!find(file))
 			 continue;
 
-		    if (!(strlen(x) < sizeof(cmd->cmd)))
-		    {
-			 fprintf(stderr, "MozPlugger: Error: Command too long (%s)\n", x);
-			 exit(1);
-		    }
-
-		    strncpy(cmd->cmd, x, sizeof(cmd->cmd));
+		    cmd->cmd = makeStrStatic(x, strlen(x));
 	       }
 	       else
 	       {
@@ -596,7 +675,9 @@ static void read_config(FILE *f)
 /*****************************************************************************
  * Find the config file or the helper file in function of the function cb.
  *****************************************************************************/
-static int find_helper_file(char *basename, int (*cb)(char *,void *data), void *data)
+static int find_helper_file(const char *basename, 
+                            int (*cb)(const char *,void *data), 
+                            void *data)
 {
      static char fname[LARGE_BUFFER_SIZE];
      char *tmp;
@@ -644,7 +725,7 @@ static int find_helper_file(char *basename, int (*cb)(char *,void *data), void *
      return 0;
 }
 
-static int read_config_cb(char *fname, void *data)
+static int read_config_cb(const char *fname, void *data)
 {
      int m4out[2];
      pid_t pid;
@@ -692,11 +773,10 @@ static int read_config_cb(char *fname, void *data)
 	  waitpid(pid, 0, 0);
 	  config_fname = strdup(fname);
      }
-
      return 1;
 }
 
-static int find_plugger_helper_cb(char *fname, void *data)
+static int find_plugger_helper_cb(const char *fname, void *data)
 {
      struct stat buf;
      if (stat(fname, &buf)) return 0;
@@ -704,7 +784,7 @@ static int find_plugger_helper_cb(char *fname, void *data)
      return 1;
 }
 
-static int find_plugger_controller_cb(char *fname, void *data)
+static int find_plugger_controller_cb(const char *fname, void *data)
 {
      struct stat buf;
      if (stat(fname, &buf)) return 0;
@@ -754,7 +834,7 @@ static void do_read_config(void)
  * Since href's are passed to an app as an argument, just check for ways that 
  * a shell can be tricked into executing a command.
  *****************************************************************************/
-static int safeURL(char* url)
+static int safeURL(const char* url)
 {
      int  i = 0; 
      int  len = strlen(url);
@@ -773,11 +853,56 @@ static int safeURL(char* url)
      return 1;
 }
 
+/*****************************************************************************
+ * See if the URL matches out match criteria.
+ *****************************************************************************/
+__inline
+static int match_url(const char * matchStr, const char * url)
+{
+     int matchStrLen;
+     const char * end;
+
+     switch (matchStr[0])
+     {
+     case '*':
+          /* Does the URL start with the match String */
+          matchStr++;     /* Step over the asterisk */
+	  return (strncasecmp(matchStr, url, strlen(matchStr)) == 0);
+
+     case '%':
+          /* Does the URL end with the match String */
+          matchStr++;     /* Step over the percent sign */
+
+          /* Need to find the end of the url, before any 
+           * extra params i.e'?=xxx' or '#yyy' */
+          end = strchr(url, '?');
+          if(end == NULL)
+          {
+               end = strchr(url, '#');
+               if(end == NULL)
+               {
+                    end = &url[strlen(url)];
+               }
+          }
+          matchStrLen = strlen(matchStr);
+          if(end - matchStrLen < url)
+          {
+               return 0;
+          }
+	  return (strncasecmp(matchStr, end-matchStrLen, matchStrLen) == 0);
+ 
+     default:
+          /* Is the match string anywhere in the URL */
+	  return (strstr(url, matchStr) != NULL);
+     }
+}
+
 /*****************************************************************************
  * Go through the commands in the config file and find one that fits our needs.
  *****************************************************************************/
 __inline
-static int match_command(NPP instance, int streaming, command_t *c)
+static int match_command(NPP instance, int streaming, 
+                         command_t *c, const char * url)
 {
 #define MODE_MASK (H_NOEMBED | H_EMBED | H_LINKS)
 
@@ -806,10 +931,22 @@ static int match_command(NPP instance, int streaming, command_t *c)
 	  return 0;
      }
 
+     if(c->fmatchStr)
+     {
+          if(!match_url(c->fmatchStr, url))
+          {
+               D("fmatch mismatch: url '%s' doesnt have '%s'\n", 
+                                              url, c->fmatchStr);
+               return 0;
+          }
+     }
      D("Match found!\n");
      return 1;
 }
 
+/*****************************************************************************
+ * See if mimetype matches.
+ *****************************************************************************/
 __inline
 static int match_mime_type(NPP instance, mimetype_t *m)
 {
@@ -829,8 +966,12 @@ static int match_mime_type(NPP instance, mimetype_t *m)
      return 1;
 }
 
+/*****************************************************************************
+ * See if handler matches.
+ *****************************************************************************/
 __inline
-static int match_handler(handler_t *h, NPP instance, int streaming)
+static int match_handler(handler_t *h, NPP instance, 
+                         int streaming, const char *url)
 {
      mimetype_t *m;
      command_t *c;
@@ -847,7 +988,7 @@ static int match_handler(handler_t *h, NPP instance, int streaming)
 	       for (cid = 0; cid < h->num_cmds; cid++)
 	       {
 		    c = &h->cmds[cid];
-		    if (match_command(instance, streaming, c))
+		    if (match_command(instance, streaming, c, url))
 		    {
 			 THIS->cmd_flags = c->flags;
 			 THIS->command = c->cmd;
@@ -860,7 +1001,10 @@ static int match_handler(handler_t *h, NPP instance, int streaming)
      return 0;
 }
 
-static int find_command(NPP instance, int streaming)
+/*****************************************************************************
+ * Find the appropriate command
+ *****************************************************************************/
+static int find_command(NPP instance, int streaming, const char * url)
 {
      int hid;
 
@@ -870,7 +1014,7 @@ static int find_command(NPP instance, int streaming)
 
      for (hid = 0; hid < num_handlers; hid++)
      {
-	  if (match_handler(&handlers[hid], instance, streaming))
+	  if (match_handler(&handlers[hid], instance, streaming, url))
 	  {
 	       D("Command found.\n");
 	       return 1;
@@ -990,7 +1134,7 @@ NPError NPP_GetValue(void *instance, NPPVariable variable, void *value)
  * Convert a string to an integer.
  * The string can be true, false, yes or no.
  *****************************************************************************/
-static int my_atoi(char *s, int my_true, int my_false)
+static int my_atoi(const char *s, int my_true, int my_false)
 {
      switch (s[0])
      {
@@ -1041,8 +1185,8 @@ NPError NPP_New(NPMIMEType pluginType,
 	  return NPERR_INVALID_INSTANCE_ERROR;
      }
 
-     THIS = NPN_MemAlloc(sizeof(data_t));
-     if (!THIS) 
+     instance->pdata = NPN_MemAlloc(sizeof(data_t));
+     if (instance->pdata == NULL) 
           return NPERR_OUT_OF_MEMORY_ERROR;
 
 
@@ -1253,8 +1397,9 @@ NPError NPP_Destroy(NPP instance, NPSavedData** save)
 	  free(THIS->mimetype);
 	  THIS->href = NULL;
 	  THIS->mms = NULL;
-	  NPN_MemFree(THIS);
-	  THIS = NULL;
+
+	  NPN_MemFree(instance->pdata);
+	  instance->pdata = NULL;
      }
 
      D("Destroy finished\n");
@@ -1367,7 +1512,8 @@ NPError NPP_NewStream(NPP instance,
      }
 
      /* Find a suitable command, fall back to savedMimeType if defined */
-     while(!find_command(instance, 1) && !find_command(instance, 0))
+     while(!find_command(instance, 1, stream->url) 
+                                 && !find_command(instance, 0, stream->url))
      {
           if(savedMimetype)       /* Lets try the saved one */
           {
@@ -1423,11 +1569,12 @@ void NPP_StreamAsFile(NPP instance,
  ****************************************************************************/
 void resize_window(NPP instance)
 {
+     XSetWindowAttributes attrib;
+
      /* Mozilla work around not require for versions > 0.13 */
      if((browserApiMajorVer > 0) || (browserApiMinorVer > 13))
          return;
 
-     XSetWindowAttributes attrib;
      attrib.override_redirect = True;
      XChangeWindowAttributes(THIS->display, (Window)THIS->windata.window,
 			     (unsigned long) CWOverrideRedirect, &attrib);
@@ -1477,7 +1624,7 @@ NPError NPP_SetWindow(NPP instance, NPWindow* window)
           {
                /* Can only use streaming commands, as Mozilla cannot handle
                 * these types (mms) of urls */
-               if (!find_command(instance, 1))
+               if (!find_command(instance, 1, THIS->mms))
                {
 	           THIS->mms = NULL;
 	           NPN_Status(instance, 
@@ -1530,6 +1677,14 @@ NPError NPP_Initialize(void)
              browserApiMajorVer, browserApiMinorVer);
 
      do_read_config();
+
+#ifdef DEBUG
+     {
+         const int free = MAX_STATIC_MEMORY_POOL - staticPoolIdx;
+         D("Static Pool used=%i, free=%i\n", staticPoolIdx, free);
+     }
+#endif
+
      return NPERR_NO_ERROR;
 }
 
